"""
CTTS

Z. Zeng, R. Kaur, S. Siddagangappa, S. Rahimi, T. Balch, and M. Veloso, “Financial Time Series Forecasting using CNN and Transformer.” arXiv, Apr. 10, 2023. Accessed: Sep. 24, 2023. [Online]. Available: http://arxiv.org/abs/2304.04912
"""

import torch
from torch import nn
from ..modules import MultiheadedSelfAttentionModule

__all__ = ["CTTS"]

class CTTS(nn.Module):
    def __init__(self,
                 window: int = 80,
                 embed_dim: int = 128,
                 kernel_size: int = 16,
                 stride: int = 8,
                 attn_depth: int = 4,
                 attn_heads: int = 4,
                 dropout: float = 0.3,
                 categories: int = 2,
                 ) -> None:
        super(CTTS, self).__init__()
        self.window = window
        self.embed_dim = embed_dim
        self.kernel_size = kernel_size
        self.stride = stride
        self.attn_depth = attn_depth
        self.attn_heads = attn_heads
        self.dropout = dropout
        self.categories = categories
        self.token_len = (window - kernel_size) // stride + 1
        # Define modules
        self.conv1d = nn.Conv1d(in_channels=1, out_channels=embed_dim, kernel_size=kernel_size, stride=stride)
        self.attns = nn.Sequential(*(MultiheadedSelfAttentionModule(num_head=attn_heads, embed_dim=embed_dim, dropout=dropout) for _ in range(attn_depth)))
        self.mlp = nn.Sequential(
            nn.Linear(self.token_len * embed_dim, self.token_len * embed_dim * 2 // 3),
            nn.GELU(),
            nn.Linear(self.token_len * embed_dim * 2 // 3, categories),
        )

    def forward(self,
                x, # [bsz, window]
                ) -> torch.Tensor:
        assert x.shape[1] == self.window, "Wrong input tensor shape."
        # Nornmalize and reshape
        x = (x - x.min(dim=1, keepdim=True).values) /                                               \
            (x.max(dim=1, keepdim=True).values - x.min(dim=1, keepdim=True).values)                 # [bsz, window]
        x = x.unsqueeze(dim=1)                                                                      # [bsz, 1, window]
        # Extract latent dependencies
        x = self.conv1d(x)                                                                          # [bsz, embed_dim, token_len]
        x = self.attns(x.transpose(1, 2))                                                           # [bsz, token_len, embed_dim]
        # Predict
        x = self.mlp(x.flatten(1, 2))                                                               # [bsz, categories]
        return x



import numpy as np
import torch
from torch import nn
from torch.utils.data import DataLoader, Dataset
from tqdm.auto import tqdm

import logging
from typing import Optional, Callable, Union

g_device = "cuda:0" if torch.cuda.is_available() else "cpu"

__all__ = [
    "VerboseBase",
    "dl_trainer",
    "dl_evaluator",
]

class VerboseBase:
    """The base for all Verbosers

    Functions
    ---------
    epoch_end: [int, str] -> Optional[float]:
        Will be called when an epoch ends, now you can summarize your results and output/log your
    evaluation, additionally, you may also return your metric to override early stopping (Note that
    the metric also assumes less being better, so metrics like accuracy should be negatived before
    returning)
    eval_iter: [torch.Tensor, torch.Tensor] -> None:
        Will be called each epoch, and you may calculate whatever metric you want

    """
    def __init__(self):
        pass

    def epoch_end(self,
                  epoch: int,
                  mode: str,
                  ) -> Optional[float]:
        raise NotImplementedError

    def eval_iter(self,
                  yhat: torch.Tensor,
                  y: torch.Tensor
                  ) -> None:
        raise NotImplementedError

class dl_trainer:
    """The torch model trainer

    Parameters
    ----------
    model: torch.nn.Module
        The model to be trained
    loss: Callable[[torch.nn.tensor, torch.nn.tensor], torch.tensor]
        The loss function
    optimizer: torch.optim.Optimizer
        The BP optimizer
    epoch: int
        Number of epochs to train
    lr: float,
        The learning rate
    lr_scheduler: Optional[Callable[[int], float]], default `None`
        A function that computes a multiplicative factor given an integer parameter epoch, `None`
    to disable
    batch_size: Union[int, Callable[[int], int]], default `64`
        A function that computes batch size given an integer parameter epoch, or a constant batch
    size, only enable when passing a
    `torch.utils.data.Dataset` to the trainer
    mini_batch: Union[int, Callable[[int], int]], default `1`
        A function that computes mini batch size given an integer parameter epoch, or a constant
    mini batch size, useful when you have a small GPU
    memory size but want much larger batch sizes
    patience: int, default `-1`:
        The patience parameter for early stopping, set to `-1` to disable
    delta: float, default `0`:
        The minimal loss improvement required for the model to be considered improving
    logger: Optional[logging.Logger], default `None`
        A logger for debug-level information
    verbose: Optional[VerboseBase], default `None`
        The verbose evaluator for epoch level data
    eval_train: bool, default `False`
        Whether to eval the train set every epoch
    device: torch.device, default `"cuda:0" if torch.cuda.is_available() else "cpu"`
        On which device to train the model
    *args, **kwargs:
        Additional args for optimizer
    """
    def __init__(self,
                 model: nn.Module,
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
                 optimizer: torch.optim.Optimizer,
                 epoch: int,
                 lr: float,
                 lr_scheduler: Optional[Callable[[int], float]] = None,
                 batch_size: Union[int, Callable[[int], int]] = 64,
                 mini_batch: Union[int, Callable[[int], int]] = 1,
                 patience: int = -1,
                 delta: float = 0,
                 logger: Optional[logging.Logger] = None,
                 verbose: Optional[VerboseBase] = None,
                 eval_train: bool = False,
                 device: torch.device = "cuda:0" if torch.cuda.is_available() else "cpu",
                 *args, **kwargs,
                 ) -> None:
        self.model = model
        self.loss = loss
        self.lr = lr
        self.optimizer = optimizer(filter(lambda layer: layer.requires_grad, self.model.parameters()), lr=lr, *args, **kwargs)
        self.epoch = epoch
        self.lr_lambda = lr_scheduler
        self.lr_scheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer, lr_lambda=lr_scheduler) if lr_scheduler is not None else None
        self.batch_size = batch_size
        self.mini_batch = mini_batch
        self.patience = patience
        self.delta = delta
        self.logger = logger
        self.verbose = verbose
        self.eval_train = eval_train
        self.device = device

    def train(self,
              train_data: Union[Dataset, DataLoader],
              valid_data: Union[Dataset, DataLoader],
              *args, **kwargs) -> None:
        """Train the model

        Parameters
        ----------
        train_data: Union[Dataset, DataLoader]
            The training data
        valid_date: Union[Dataset, DataLoader]
            The validation data
        *args, **kwargs:
            Additional parameters for `torch.utils.data.DataLoader` if passing a `torch.utils.data.Dataset`
        """
        self.logger.info("Begin training model. Training params: loss = {}; optimizer = {}.".format(self.loss, self.optimizer)) if self.logger else None # INFO-LOG
        torch.cuda.empty_cache()
        self.model.to(self.device)
        best_loss, best_metric, patience = np.inf, np.inf, self.patience
        self.model.to(self.device)
        for EPOCH in tqdm(range(self.epoch), desc="training", unit="epoch"):
            self.logger.info("Start training@EPOCH {}. Current params: lr = {}{}{}; patience = {}/{}".format(EPOCH, self.lr, " * {}".format(self.lr_lambda(EPOCH)) if self.lr_lambda else "", "; batch size = {} * {}".format(self.batch_size(EPOCH) if callable(self.batch_size) else self.batch_size, self.mini_batch(EPOCH) if callable(self.mini_batch) else self.mini_batch) if isinstance(train_data, Dataset) else "", patience, self.patience)) if self.logger else None # INFO-LOG
            if patience > 0:
                patience = patience - 1
            self.train_epoch(EPOCH, train_data, *args, **kwargs)
            if self.eval_train:
                loss = self.eval_epoch(EPOCH, train_data, *args, **kwargs)
                self.logger.info(f"Finished train-eval@EPOCH {EPOCH}, current loss = {loss:.8f}") if self.logger else None # INFO-LOG
                if self.verbose is not None:
                    metric = self.verbose.epoch_end(EPOCH, "train")
            loss = self.eval_epoch(EPOCH, valid_data, *args, **kwargs)
            if loss < best_loss - self.delta:
                self.logger.info(f"Finished validing@EPOCH {EPOCH}, current loss = {loss:.8f}, loss update {best_loss:.8f} -> {loss:.8f}") if self.logger else None # INFO-LOG
                best_loss = loss
                patience = self.patience
            else:
                self.logger.info(f"Finished validing@EPOCH {EPOCH}, current loss = {loss:.8f}") if self.logger else None # INFO-LOG
            if self.verbose is not None:
                metric = self.verbose.epoch_end(EPOCH, "valid")
                if metric is not None and metric < best_metric:
                    self.logger.info(f"Verbose metric: {metric:.8f}, metric update {best_metric:.8f} -> {metric:.8f}.") if self.logger else None # INFO LOG
                    best_metric = metric
                    patience = self.patience
                else:
                    self.logger.info(f"Verbose metric: {metric:.8f}.") if self.logger else None # INFO LOG
            if patience == 0:
                break

            if self.lr_scheduler is not None:
                self.lr_scheduler.step()

    def train_epoch(self,
                    epoch: int,
                    train_data: Union[Dataset, DataLoader],
                    *args, **kwargs
                    ) -> None:
        if isinstance(train_data, DataLoader):
            train_loader = train_data
        elif isinstance(self.batch_size, int):
            train_loader = DataLoader(train_data, batch_size=self.batch_size, *args, **kwargs)
        elif callable(self.batch_size):
            train_loader = DataLoader(train_data, batch_size=self.batch_size(epoch), *args, **kwargs)
        else:
            raise ValueError
        mini_batch, countr = self.mini_batch if isinstance(self.mini_batch, int) else self.mini_batch(epoch), 0
        self.model.train()
        for idx, packet in enumerate(train_loader):
            x, y = packet[0].to(self.device), packet[1].to(self.device)
            yhat = self.model(x)
            loss = self.loss(yhat, y) / mini_batch
            if countr == 0:
                self.optimizer.zero_grad()
            loss.backward()
            countr = countr + 1
            if countr == mini_batch:
                self.optimizer.step()
                self.logger.debug("Training iter {}, loss = {:.8f}, optimizer stepped.".format(idx, loss.data * mini_batch)) if self.logger else None # DEBUG-LOG
                countr = 0
            else:
                self.logger.debug("Training iter {}, loss = {:.8f}.".format(idx, loss.data * mini_batch)) if self.logger else None # DEBUG-LOG

    def eval_epoch(self,
                   epoch: int,
                   valid_data: Union[Dataset, DataLoader],
                   *args, **kwargs
                   ) -> float:
        if isinstance(valid_data, DataLoader):
            valid_loader = valid_data
        elif isinstance(self.batch_size, int):
            valid_loader = DataLoader(valid_data, batch_size=self.batch_size, *args, **kwargs)
        elif callable(self.batch_size):
            valid_loader = DataLoader(valid_data, batch_size=self.batch_size(epoch), *args, **kwargs)
        else:
            raise ValueError
        accum_loss, accum_batch = 0, 0
        self.model.eval()
        with torch.no_grad():
            for x, y in valid_loader:
                x, y = x.to(self.device), y.to(self.device)
                yhat = self.model(x)
                loss = self.loss(yhat, y)
                accum_loss = accum_loss + loss.data * y.shape[0]
                accum_batch = accum_batch + y.shape[0]
                if self.verbose is not None:
                    self.verbose.eval_iter(yhat=yhat, y=y)
        return accum_loss / accum_batch

class dl_evaluator:
    """The torch model evaluator

    Parameters
    ----------
    model: torch.nn.Module
        The trained model.
    verbose: Optional[VerboseBase], default `None`
        The verbose evaluator for infomation output
    device: torch.device, default `"cuda:0" if torch.cuda.is_available() else "cpu"`
        On which device to evaluate the model
    """
    def __init__(self,
                 model: nn.Module,
                 verbose: Optional[VerboseBase] = None,
                 device: torch.device = "cuda:0" if torch.cuda.is_available() else "cpu",
                 ) -> None:
        self.model = model
        self.verbose = verbose
        self.device = device

    def eval(self,
             valid_data: Union[Dataset, DataLoader],
             *args, **kwargs,
             ) -> float:
        if isinstance(valid_data, DataLoader):
            valid_loader = valid_data
        else:
            valid_loader = DataLoader(valid_data, *args, **kwargs)
        self.model.eval()
        with torch.no_grad():
            for x, y in valid_loader:
                x, y = x.to(self.device), y.to(self.device)
                yhat = self.model(x)
                if self.verbose is not None:
                    self.verbose.eval_iter(yhat=yhat, y=y)
        self.verbose.epoch_end(0, "eval")



import torch
from torch import nn
from .transformers import FeedForwardLayer

from typing import Optional, Callable

class MixtureOfExpertsBlock(nn.Module):
    """The Mixture of Experts Module
    - MoE
      N. Shazeer et al., “Outrageously Large Neural Networks: The Sparsely-Gated Mixture-of-Experts
      Layer.” arXiv, Jan. 23, 2017. Accessed: Sep. 27, 2023. [Online]. Available:
      http://arxiv.org/abs/1701.06538
    - FastMoE
      J. He, J. Qiu, A. Zeng, Z. Yang, J. Zhai, and J. Tang, “FastMoE: A Fast Mixture-of-Expert Tra
      ining System.” arXiv, Mar. 24, 2021. Accessed: Sep. 28, 2023. [Online]. Available:
      http://arxiv.org/abs/2103.13262
    - TODO: ST-MoE
    B. Zoph et al., “ST-MoE: Designing Stable and Transferable Sparse Expert Models.” arXiv, Apr.
    29, 2022. Accessed: Sep. 27, 2023. [Online]. Available: http://arxiv.org/abs/2202.08906

    Parameters
    ----------
    embed_dim: int
        the dimension of input embeddings
    num_experts: int
        the number of expert modules
    topk: int, default `2`
        the number of topk experts to activate
    gate: Optional[Callable[[torch.Tensor], torch.Tensor]], default `None`
        the gate layer, if `None` default to simple linear transformation
    expert: torch.nn.Module, default `sifeng.dl.FeedForwardLayer`
        the choice of expert module
    *args, **kwargs:
        parameters for your choice of the expert module

    Input shape
    -----------
    [bsz * ?? * embed_dim]

    """
    def __init__(self,
                 embed_dim: int,
                 num_experts: int,
                 topk: int = 2,
                 gate: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
                 expert: nn.Module = FeedForwardLayer,
                 *args, **kwargs) -> None:
        super(MixtureOfExpertsBlock, self).__init__()
        self.embed_dim = embed_dim
        self.num_experts = num_experts
        self.topk = topk

        self.softmax = nn.Softmax(dim=-1)
        self.gate = gate if gate is not None else nn.Linear(embed_dim, self.num_experts)
        self.experts = nn.ModuleList([expert(embed_dim=embed_dim, *args, **kwargs) for _ in range(self.num_experts)])

    # TODO: Functionize MoE module.
    def forward(self,
                x: torch.Tensor, # [bsz * ?? * embed_dim]
                ) -> torch.Tensor:
        device = x.device
        fsz = x.shape[:-1] # fsz = [bsz * ??]
        x = x.flatten(start_dim=0, end_dim=-2) # [fsz * embed_dim]
        gate = self.gate(x) # [fsz * num_experts]
        topk_values, topk_indices = torch.topk(gate, k=self.topk, dim=-1, sorted=True) # [fsz * topk], [fsz * topk]
        topk_weight = self.softmax(topk_values) # [fsz * topk]
        mask = torch.zeros([topk_weight.shape[0], self.num_experts], device=device, dtype=topk_weight.dtype).scatter(dim=1, index=topk_indices, src=topk_weight) # [fsz * num_experts]
        activators = [mask[:, i].nonzero().squeeze(-1) for i in range(self.num_experts)] # List([?], len=num_experts)
        weights = [mask[:, i].index_select(dim=0, index=activators[i]) for i in range(self.num_experts)]
        inputs = [x.index_select(dim=0, index=activators[i]) for i in range(self.num_experts)] # List([? * embed_dim], len=num_experts)
        outputs = [torch.mul(self.experts[i](inputs[i]), weights[i].unsqueeze(dim=-1)) for i in range(self.num_experts)] # List([? * embed_dim], len=num_experts)
        x = torch.zeros(x.shape[0], outputs[0].shape[-1], device=device, dtype=x.dtype) # [fsz * embed_dim]
        for i in range(self.num_experts):
            x = x.scatter_add(0, activators[i].unsqueeze(dim=-1).repeat_interleave(outputs[0].shape[-1], dim=-1), outputs[i]) # [fsz * embed_dim]
        x = x.unflatten(dim=0, sizes=fsz) # [bsz * ?? * embed_dim]
        return x



import torch
from .module import BayesianModule

__all__ = [
    "Sequential",
]

class Sequential(BayesianModule):
    """A sequential container

    The same as `torch.nn.Sequential`, except that bayesian modules inside will automatically calcu
    late their prior and posterior distributions.

    Parameters
    ----------
    *args:
        The sequence of `BayesianModules`

    """
    def __init__(self, *args):
        super(Sequential, self).__init__()
        self.sequence = torch.nn.Sequential(*args)
        dict = self.__dict__.get("_bmodules")
        for idx, module in enumerate(args):
            if isinstance(module, BayesianModule):
                dict[str(idx)] = module

    def forward(self, x):
        return self.sequence(x)



from .transformers import FeedForwardLayer, MultiheadedSelfAttentionModule, TransformerEncoderBlock
from .positional_encoding import AbsoluteSinusoidalPE, AbsoluteLearnablePE
from .sparse_model import MixtureOfExpertsBlock

__all__ = [
    "FeedForwardLayer", "MultiheadedSelfAttentionModule", "TransformerEncoderBlock",
    "AbsoluteSinusoidalPE", "AbsoluteLearnablePE", "MixtureOfExpertsBlock",
]



import torch
from torch.nn.modules.module import Module
from collections import OrderedDict
from typing import Any

__all__ = [
    "BayesianModule",
]

class BayesianModule(Module):
    """Base class for all bayesian neural network modules.

    Your bayesian models should also subclass this class. Be sure to override `forward()`.

    """
    def __init__(self):
        super(BayesianModule, self).__init__()
        super(BayesianModule, self).__setattr__('_bmodules', OrderedDict())
        self.register_buffer("log_prior", torch.tensor(0.0))
        self.register_buffer("log_posterior", torch.tensor(0.0))

    def __getattr__(self, name: str) -> Any:
        return super(BayesianModule, self).__getattr__(name)

    def __setattr__(self, name: str, value: Any) -> None:
        if isinstance(value, BayesianModule):
            bmodules = self.__dict__.get("_bmodules")
            bmodules[name] = value
        super(BayesianModule, self).__setattr__(name, value)

    def bmodules(self):
        return self.__dict__.get("_bmodules").values()

    def _log_prior(self):
        log_prior = self.log_prior
        for bmodule in self.bmodules():
            log_prior = log_prior + bmodule._log_prior()
        return log_prior

    def _log_posterior(self):
        log_posterior = self.log_posterior
        for bmodule in self.bmodules():
            log_posterior = log_posterior + bmodule._log_posterior()
        return log_posterior



import torch, duckdb, requests, pandas as pd
from tqdm.auto import tqdm
from pathlib import Path
from datetime import datetime
from joblib import Parallel, delayed
from pandas.tseries.offsets import MonthEnd
from math import floor, ceil
from colorama import Fore, Back, Style

from typing import Union, Literal, List

__all__ = [
    "stat",
    "basic_info", "kline_day", "indicator_day",
]

def stat(model: torch.nn.Module,
         sort: Literal[":idx", "idx:", ":name", "name:", ":train", "train:", ":params", "params:", ":memory", "memory:"] = "train:",
         ) -> str:
    """Prints analysis for a specific model

    Parameters
    ----------
    model: torch.nn.Module
        the model to be interepted
    sort: Literal[":idx", "idx:", ":name", "name:", ":train", "train:", ":params", "params:", ":memory", "memory:"], deafult `"train:"`
        how to sort the output, a leading comma means ascending and a trailing comma means descending
    """
    def fmt(info, length, place):
        info = str(info)
        if place == "center":
            return " " * floor((length - len(info)) / 2) + info + " " * ceil((length - len(info)) / 2)
        elif place == "left":
            return " " + info + " " * floor(length - len(info) - 1)
        elif place == "right":
            return " " * floor(length - len(info) - 1) + info + " "
    def elafmt(sz, carry, seq, dig):
        idx = 0
        while sz >= carry and idx < len(seq) - 1:
            sz = sz / carry
            idx = idx + 1
        if idx == 0 and seq[idx] == "":
            return "{:d}".format(int(sz))
        elif idx == 0 or int(sz) == sz:
            return "{:d} {}".format(int(sz), seq[idx])
        elif dig == 2:
            return "{:.2f} {}".format(sz, seq[idx])
        elif dig == 4:
            return "{:.4f} {}".format(sz, seq[idx])
    mem_lookup = {
        torch.uint8: 1,
        torch.ByteTensor: 1,
        torch.int8: 1,
        torch.CharTensor: 1,
        torch.int16: 2,
        torch.short: 2,
        torch.ShortTensor: 2,
        torch.float16: 2,
        torch.half: 2,
        torch.HalfTensor: 2,
        torch.int32: 4,
        torch.int: 4,
        torch.IntTensor: 4,
        torch.float: 4,
        torch.float32: 4,
        torch.FloatTensor: 4,
        torch.float64: 8,
        torch.double: 8,
        torch.DoubleTensor: 8,
        torch.int64: 8,
        torch.long: 8,
        torch.LongTensor: 8
    }

    assert isinstance(model, torch.nn.Module), "model is not a torch module."

    info = ""
    model_info, model_params, model_memsz = [], 0, 0
    row_width = (
        6 if sort in [":idx", "idx:"] else 5,
        7 if sort in [":name", "name:"] else 6,
        7,
        8 if sort in [":train", "train:"] else 7,
        9 if sort in [":params", "params:"] else 8,
        9 if sort in [":memory", "memory:"] else 8,
    )
    for idx, (name, w_variable) in enumerate(model.named_parameters()):
        shape = str(w_variable.shape)
        params = 1
        for dim_size in w_variable.shape:
            params = params * dim_size
        model_params = model_params + params
        model_memsz = model_memsz + mem_lookup[w_variable.dtype] * params
        train = "True" if w_variable.requires_grad else "False"
        row_width = (max(row_width[0], len(str(idx)) + 2),
                     max(row_width[1], len(str(name)) + 2),
                     max(row_width[2], len(str(shape)) + 2),
                     max(row_width[3], len(train) + 2),
                     max(row_width[4], len(elafmt(params, 1000, ["", "K", "M", "B", "T"], 2)) + 2),
                     max(row_width[5], len(elafmt(mem_lookup[w_variable.dtype] * params, 1024, ["Bytes", "KiB", "MiB", "GiB", "TiB"], 2)) + 2),
                     )
        model_info.append((idx, name, shape, train, params, mem_lookup[w_variable.dtype] * params))

    assert sort in [":idx", "idx:", ":name", "name:", ":train", "train:", ":params", "params:", ":memory", "memory:"], \
        "Unsupported sorting direction."
    if sort == ":idx":
        model_info.sort(key=lambda x: (x[0]), reverse=False)
    elif sort == "idx:":
        model_info.sort(key=lambda x: (x[0]), reverse=True)
    elif sort == ":name":
        model_info.sort(key=lambda x: (x[1], x[0]), reverse=False)
    elif sort == "name:":
        model_info.sort(key=lambda x: (x[1], -x[0]), reverse=True)
    elif sort == ":train":
        model_info.sort(key=lambda x: (x[3], x[0]), reverse=False)
    elif sort == "train:":
        model_info.sort(key=lambda x: (x[3], -x[0]), reverse=True)
    elif sort == ":params":
        model_info.sort(key=lambda x: (x[4], x[0]), reverse=False)
    elif sort == "params:":
        model_info.sort(key=lambda x: (x[4], -x[0]), reverse=True)
    elif sort == ":memory":
        model_info.sort(key=lambda x: (x[5], x[0]), reverse=False)
    elif sort == "memory:":
        model_info.sort(key=lambda x: (x[5], -x[0]), reverse=True)

    info = info + "┌" + "─" * row_width[0] + \
                  "┬" + "─" * row_width[1] + \
                  "┬" + "─" * row_width[2] + \
                  "┬" + "─" * row_width[3] + \
                  "┬" + "─" * row_width[4] + \
                  "┬" + "─" * row_width[5] + "┐\n"
    info = info + "│{}│{}│{}│{}│{}│{}│\n".format(
        fmt(sort if sort in [":idx", "idx:"] else "idx", row_width[0], "center"),
        fmt(sort if sort in [":name", "name:"] else "name", row_width[1], "center"),
        fmt("shape", row_width[2], "center"),
        fmt(sort if sort in [":train", "train:"] else "trainable", row_width[3], "center"),
        fmt(sort if sort in [":params", "params:"] else "params", row_width[4], "center"),
        fmt(sort if sort in [":memory", "memory:"] else "memory", row_width[5], "center"),
    )
    info = info + "├" + "─" * row_width[0] + \
                  "┼" + "─" * row_width[1] + \
                  "┼" + "─" * row_width[2] + \
                  "┼" + "─" * row_width[3] + \
                  "┼" + "─" * row_width[4] + \
                  "┼" + "─" * row_width[5] + "┤\n"
    if len(model_info) != 0:
        for item in model_info:
            if item[3] == "True":
                info = info + "│{}{}│{}│{}│{}│{}│{}{}│\n".format(
                    Style.NORMAL,
                    Fore.GREEN + fmt(item[0], row_width[0], "right") + Fore.RESET,
                    Fore.GREEN + fmt(item[1], row_width[1], "left") + Fore.RESET,
                    Fore.GREEN + fmt(item[2], row_width[2], "left") + Fore.RESET,
                    Fore.GREEN + fmt(item[3], row_width[3], "left") + Fore.RESET,
                    Fore.GREEN + fmt(elafmt(item[4], 1000, ["", "K", "M", "B", "T"], 2), row_width[4], "left") + Fore.RESET,
                    Fore.GREEN + fmt(elafmt(item[5], 1024, ["Bytes", "KiB", "MiB", "GiB", "TiB"], 2), row_width[5], "left") + Fore.RESET,
                    Fore.RESET + Back.RESET + Style.RESET_ALL,
                )
            elif item[3] == "False":
                info = info + "│{}{}│{}│{}│{}│{}│{}{}│\n".format(
                    Style.DIM,
                    Fore.RED + fmt(item[0], row_width[0], "right") + Fore.RESET,
                    Fore.RED + fmt(item[1], row_width[1], "left") + Fore.RESET,
                    Fore.RED + fmt(item[2], row_width[2], "left") + Fore.RESET,
                    Fore.RED + fmt(item[3], row_width[3], "left") + Fore.RESET,
                    Fore.RED + fmt(elafmt(item[4], 1000, ["", "K", "M", "B", "T"], 2), row_width[4], "left") + Fore.RESET,
                    Fore.RED + fmt(elafmt(item[5], 1024, ["Bytes", "KiB", "MiB", "GiB", "TiB"], 2), row_width[5], "left"),
                    Fore.RESET + Back.RESET + Style.RESET_ALL,
                )
    else:
        info = info + "│{}│\n".format(fmt("No Parameters", row_width[0] + row_width[1] + row_width[2] + row_width[3] + row_width[4] + row_width[5] + 5, "center"))
    info = info + "├" + "─" * row_width[0] + \
                  "┴" + "─" * row_width[1] + \
                  "┴" + "─" * row_width[2] + \
                  "┴" + "─" * row_width[3] + \
                  "┴" + "─" * row_width[4] + \
                  "┴" + "─" * row_width[5] + "┤\n"
    info = info + "│{}│\n".format(fmt("Model params: {}".format(elafmt(model_params, 1000, ["", "K", "M", "B", "T"], 4)), row_width[0] + row_width[1] + row_width[2] + row_width[3] + row_width[4] + row_width[5] + 5, "left"))
    info = info + "│{}│\n".format(fmt("Model memory: {}".format(elafmt(model_memsz, 1024, ["Bytes", "KiB", "MiB", "GiB", "TiB"], 4)), row_width[0] + row_width[1] + row_width[2] + row_width[3] + row_width[4] + row_width[5] + 5, "left"))
    info = info + "└" + "─" * (row_width[0] + row_width[1] + row_width[2] + row_width[3] + row_width[4] + row_width[5] + 5) + "┘"

    return info

baseurl = "https://data.sifeng.site/"

def basic_info(fields: Union[Literal["*"], List[str]] = "*",
               stock_code: Union[Literal["*"], List[str]] = "*",
               local_dir: Path = Path.home() / ".sifeng/parquet/status_charts/",
               force_update: bool = False):
    """Fetch basic info from server

    Parameters
    ----------
    fields: Union[Literal["*"], List[str]], default `'*'`
        The fields you want, you may choose from `['stock_code', 'stock_name', 'area', 'industry',\
    'sector', 'list_status', 'list_date', 'st_flag']`, or enter `'*'` to choose all.
    stock_code: Union[Literal["*"], List[str]], default `'*'`
        The code of the stocks you want to query, or enter `'*'` for all.
    local_dir: pathlib.Path, default `Path.home() / ".sifeng/parquet/status_charts/"`
        The local path you want to data to download to, please do not change under normal circumsta
    nces.
    force_update: bool, default `False`
        Whether to force fetching from remote server. If `False`, then will only fetch if hadn't
    fetched before.
    """
    local_dir.mkdir(parents=True, exist_ok=True)
    filename = "BASIC-INFO.parquet"
    update = not (local_dir / filename).exists()
    if force_update:
        update = True
    if update:
        resp = requests.get(baseurl + filename)
        with open(local_dir / filename, "wb") as file:
            file.write(resp.content)
    if fields == "*":
        fields = ['stock_code', 'stock_name', 'area', 'industry', 'sector', 'list_status', 'list_date', 'st_flag']
    if stock_code == "*":
        sql = f"SELECT {', '.join(fields)} FROM read_parquet('{local_dir / 'BASIC-INFO.parquet'}')"
    elif isinstance(stock_code, str):
        sql = f"SELECT {', '.join(fields)} FROM read_parquet('{local_dir / 'BASIC-INFO.parquet'}') WHERE stock_code = '{stock_code}'"
    else:
        connector = "', '"
        sql = f"SELECT {', '.join(fields)} FROM read_parquet('{local_dir / 'BASIC-INFO.parquet'}') WHERE stock_code IN ('{connector.join(stock_code)}')"
    return duckdb.sql(sql).df()

def kline_day(fields: Union[Literal["*"], List[str]] = "*",
              begin_date: str = "1999-12-19",
              end_date: str = datetime.now().strftime("%Y-%d-%m"),
              stock_code: Union[Literal["*"], List[str]] = "*",
              local_dir: Path = Path.home() / ".sifeng/parquet/kline_day/",
              n_jobs: int = 1):
    """Fetch kline (freq. day) from server

    Parameters
    ----------
    fields: Union[Literal["*"], List[str]], default `'*'`
        The fields you want, you may choose from `['stock_code', 'trade_date', 'open', 'high',\
    'low', 'close', 'vol', 'amount', 'adj_factor']`, or enter `'*'`
    to choose all.
    begin_date: str, deafult `"1999-12-19"`
        The date youe want the query to begin.
    end_date: str, default `datetime.now().strftime("%Y-%d-%m")`
        The date you want the query to end.
    stock_code: Union[Literal["*"], List[str]], default `'*'`
        The code of the stocks you want to query, or enter `'*'` for all.
    local_dir: pathlib.Path, default `Path.home() / ".sifeng/parquet/kline_day/"`
        The local path you want to data to download to, please do not change under normal circumsta
    nces.
    n_jobs: int, default `1`
        The number of workers for fetching.
    """
    def fdownload(mend):
        filename = f"KLINE-DAY{mend.strftime('%Y%m')}.parquet"
        update = not (local_dir / filename).exists()
        if mend.year == datetime.now().year and mend.month == datetime.now().month:
            update = True
        if update:
            resp = requests.get(baseurl + filename)
            with open(local_dir / filename, "wb") as file:
                file.write(resp.content)
    begin_date, end_date = pd.to_datetime(begin_date), pd.to_datetime(end_date)
    tasks = tqdm([delayed(fdownload)(_) for _ in pd.date_range(begin_date, end_date + MonthEnd(0), freq="M")], desc="Checking", unit="month", unit_scale=True, leave=False)
    local_dir.mkdir(parents=True, exist_ok=True)
    Parallel(n_jobs=n_jobs, verbose=0)(tasks)
    if fields == "*":
        fields = ['stock_code', 'trade_date', 'open', 'high', 'low', 'close', 'vol', 'amount', 'adj_factor']
    if stock_code == "*":
        sql = f"SELECT {', '.join(fields)} FROM read_parquet('{local_dir / '*.parquet'}') WHERE trade_date BETWEEN '{begin_date.strftime('%Y-%m-%d')}' AND '{end_date.strftime('%Y-%m-%d')}'"
    elif isinstance(stock_code, str):
        sql = f"SELECT {', '.join(fields)} FROM read_parquet('{local_dir / '*.parquet'}') WHERE trade_date BETWEEN '{begin_date.strftime('%Y-%m-%d')}' AND '{end_date.strftime('%Y-%m-%d')}' AND stock_code = '{stock_code}'"
    else:
        connector = "', '"
        sql = f"SELECT {', '.join(fields)} FROM read_parquet('{local_dir / '*.parquet'}') WHERE trade_date BETWEEN '{begin_date.strftime('%Y-%m-%d')}' AND '{end_date.strftime('%Y-%m-%d')}' AND stock_code IN ('{connector.join(stock_code)}')"
    return duckdb.sql(sql).df()


def indicator_day(fields: Union[Literal["*"], List[str]] = "*",
                  begin_date: str = "1999-12-19",
                  end_date: str = datetime.now().strftime("%Y-%d-%m"),
                  stock_code: Union[Literal["*"], List[str]] = "*",
                  local_dir: Path = Path.home() / ".sifeng/parquet/indicator_day/",
                  n_jobs: int = 1):
    """Fetch indicator (freq. day) from server

    Parameters
    ----------
    fields: Union[Literal["*"], List[str]], default `'*'`
        The fields you want, you may choose from `["stock_code", "trade_date", "turnover_rate", \
    "turnover_rate_free", "volume_ratio", "pe", "pe_ttm", "pb", "ps", "ps_ttm", "dv_ratio", \
    "dv_ttm", "total_share", "float_share", "free_share", "total_mv", "circ_mv"]`, or enter `'*'`
    to choose all.
    begin_date: str, deafult `"1999-12-19"`
        The date youe want the query to begin.
    end_date: str, default `datetime.now().strftime("%Y-%d-%m")`
        The date you want the query to end.
    stock_code: Union[Literal["*"], List[str]], default `'*'`
        The code of the stocks you want to query, or enter `'*'` for all.
    local_dir: pathlib.Path, default `Path.home() / ".sifeng/parquet/indicator_day/"`
        The local path you want to data to download to, please do not change under normal circumsta
    nces.
    n_jobs: int, default `1`
        The number of workers for fetching.
    """
    def fdownload(mend):
        filename = f"INDICATOR-DAY{mend.strftime('%Y%m')}.parquet"
        update = not (local_dir / filename).exists()
        if mend.year == datetime.now().year and mend.month == datetime.now().month:
            update = True
        if update:
            resp = requests.get(baseurl + filename)
            with open(local_dir / filename, "wb") as file:
                file.write(resp.content)
    begin_date, end_date = pd.to_datetime(begin_date), pd.to_datetime(end_date)
    tasks = tqdm([delayed(fdownload)(_) for _ in pd.date_range(begin_date, end_date + MonthEnd(0), freq="M")], desc="Checking", unit="month", unit_scale=True, leave=False)
    local_dir.mkdir(parents=True, exist_ok=True)
    Parallel(n_jobs=n_jobs, verbose=0)(tasks)
    if fields == "*":
        fields = ['stock_code', 'trade_date', 'turnover_rate', 'turnover_rate_free', 'volume_ratio', 'pe', 'pe_ttm', 'pb', 'ps', 'ps_ttm', 'dv_ratio', 'dv_ttm', 'total_share', 'float_share', 'free_share', 'total_mv', 'circ_mv']
    if stock_code == "*":
        sql = f"SELECT {', '.join(fields)} FROM read_parquet('{local_dir / '*.parquet'}') WHERE trade_date BETWEEN '{begin_date.strftime('%Y-%m-%d')}' AND '{end_date.strftime('%Y-%m-%d')}'"
    elif isinstance(stock_code, str):
        sql = f"SELECT {', '.join(fields)} FROM read_parquet('{local_dir / '*.parquet'}') WHERE trade_date BETWEEN '{begin_date.strftime('%Y-%m-%d')}' AND '{end_date.strftime('%Y-%m-%d')}' AND stock_code = '{stock_code}'"
    else:
        connector = "', '"
        sql = f"SELECT {', '.join(fields)} FROM read_parquet('{local_dir / '*.parquet'}') WHERE trade_date BETWEEN '{begin_date.strftime('%Y-%m-%d')}' AND '{end_date.strftime('%Y-%m-%d')}' AND stock_code IN ('{connector.join(stock_code)}')"
    return duckdb.sql(sql).df()



import numpy as np
import torch
from torch import nn
from torch.distributions import Normal, Categorical
from . import BayesianModule

__all__ = [
    "BayesianLoss",
    "ContinuousELBOLoss",
    "DiscreteELBOLoss",
]

class BayesianLoss(nn.Module):
    """The Bayesian Loss Function

    The Bayesian loss compares the prior and posterior distribution of the parameters in the model
    to help bayesian models diffuse when faced with OOS data.

    Parameters
    ----------
    model: BayesianModule
        Youe model, be sure to pass an instance rather than the class

    """
    def __init__(self,
                 model: BayesianModule,
                 ) -> None:
        super(BayesianLoss, self).__init__()
        self.model = model

    def forward(self) -> torch.Tensor:
        return self.model._log_posterior() - self.model._log_prior()

class ContinuousELBOLoss(nn.Module):
    """The ELBO loss for regression (Continuous) tasks

    Parameters
    ----------
    noise: float, default `0.01`
        The tolerable noise level, or namely, the std in the gaussian mixture model for approximati
    ng the distribution function.

    """
    def __init__(self,
                 noise: float = 0.01,
                 ) -> None:
        super(ContinuousELBOLoss, self).__init__()
        self.noise = noise

    def forward(self, yhat, y) -> torch.Tensor:
        logp = Normal(yhat, self.noise).log_prob(y)
        return -torch.sum(torch.logsumexp(logp, dim=0) - np.log(logp.shape[0]))

class DiscreteELBOLoss(nn.Module):
    """The ELBO loss for classification (Discrete) tasks

    """
    def __init__(self,
                 ) -> None:
        super(DiscreteELBOLoss, self).__init__()

    def forward(self, yhat, y) -> torch.Tensor:
        logp = Categorical(logits=yhat).log_prob(y.squeeze(-1))
        return -torch.sum(torch.logsumexp(logp, dim=0) - np.log(logp.shape[0]))



import numpy as np
import torch
from torch.nn import Parameter, init
from torch.nn.functional import softplus
from torch.distributions import Normal
from .module import BayesianModule

__all__ = [
    "Linear",
]

class Linear(BayesianModule):
    """The bayesian linear transformation.

    Applies a linear transformation to the incoming data :math:`y = x\mathbf{A}^T + \mathbf{b}`,
    where :math:`\mathbf{A}` and :math:`\mathbf{b}` are random variables.

    Parameters
    ----------
    in_features: int
        The size of each input sample.
    out_features: int
        The size of each output sample.
    bias: bool, default `True`
        If set to `False`, the layer will not learn an additive bias.
    eps: float, default `1e-5`
        The minimal epsilon added when calculating the standard deviations.

    Input shape
    -----------
    [sample * ?? * in_features]

    """
    def __init__(self,
                 in_features: int,
                 out_features: int,
                 bias: bool = True,
                 prior_var: float = 10.0,
                 eps: float = 1e-5,
                 ) -> None:
        super(Linear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.bias = bias
        self.prior_var = prior_var
        self.eps = eps

        self.weight_mean = Parameter(torch.empty(in_features, out_features))
        self.weight_rho = Parameter(torch.zeros(in_features, out_features))
        if bias:
            self.bias_mean = Parameter(torch.empty(out_features))
            self.bias_rho = Parameter(torch.zeros(out_features))
        self.reset_parameters()

    def reset_parameters(self,
                         ) -> None:
        init.kaiming_uniform_(self.weight_mean, a=np.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight_mean)
            bound = 1 / np.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(self.bias_mean, -bound, bound)

    def forward(self,
                x: torch.Tensor, # [sample * ?? * in_features]
                ) -> torch.Tensor:
        sample, shape = x.shape[0], x.shape[1:-1]
        x = x.flatten(1, -2) # [sample * full_shape * in_features]
        weight_sigma = softplus(self.weight_rho) + self.eps # [out_features * in_features]
        weight = torch.randn([sample, self.in_features, self.out_features], device=x.device) * weight_sigma + self.weight_mean # [sample * out_features * in_features]
        if self.bias:
            bias_sigma = softplus(self.bias_rho) + self.eps # [out_features]
            bias = torch.randn([sample, self.out_features], device=x.device) * bias_sigma + self.bias_mean # [sample * out_features]

        if self.bias:
            if self.training:
                self.log_prior = torch.sum(torch.logsumexp(Normal(0, self.prior_var).log_prob(weight), dim=0) - np.log(weight.shape[0])) + \
                                 torch.sum(torch.logsumexp(Normal(0, self.prior_var).log_prob(bias), dim=0) - np.log(bias.shape[0]))
                self.log_posterior = torch.sum(torch.logsumexp(Normal(self.weight_mean, weight_sigma).log_prob(weight), dim=0) - np.log(weight.shape[0])) + \
                                     torch.sum(torch.logsumexp(Normal(self.bias_mean, bias_sigma).log_prob(bias), dim=0) - np.log(bias.shape[0]))
            x = torch.bmm(x, weight) + bias.unsqueeze(1) # [sample * full_shape * out_features]
            return x.unflatten(1, shape)
        else:
            if self.training:
                self.log_prior = torch.sum(torch.logsumexp(Normal(0, self.prior_var).log_prob(weight), dim=0) - np.log(weight.shape[0]))
                self.log_posterior = torch.sum(torch.logsumexp(Normal(self.weight_mean, weight_sigma).log_prob(weight), dim=0) - np.log(bias.shape[0]))
            x = torch.bmm(x, weight) # [sample * full_shape * out_features]
            return x.unflatten(1, shape)



import torch
from torch.nn import Module
import torch.nn.functional as F

from typing import Optional

__all__ = [
    "FocalLoss",
]

class FocalLoss(Module):
    """An implementation of focal loss.
    The focal loss is computated by:

    .. math::
        Loss(x, class) = - \alpha (1-\mathrm{softmax}(x)[class])^{\gamma} \log(\mathrm{softmax}(x)[class])

    T.-Y. Lin, P. Goyal, R. Girshick, K. He, and P. Dollár, “Focal Loss for Dense Object
    Detection.” arXiv, Feb. 07, 2018. Accessed: Oct. 17, 2023. [Online]. Available:
    http://arxiv.org/abs/1708.02002

    Parameters
    ----------
    alpha: Optional[torch.Tensor], default `None`:
        Weighting factors in range (0,1) to balance positive vs negative examples. Only one between
    `alpha` and `freq` can be passed.
    freq: Optional[torch.Tensor], default `None`:
        The frequency which each category appears. Only one between `alpha` and `freq` can be passed.
    gamma: float, default `2.0`
        Exponent of the modulating factor (1 - p_t) to balance easy vs hard examples.
    reduction: str, default `mean`
        - `none`: No reduction will be applied to the output.
        - `mean`: The output will be averaged.
        - `sum`: The output will be summed.
    """
    def __init__(self,
                 alpha: Optional[torch.Tensor] = None,
                 freq: Optional[torch.Tensor] = None,
                 gamma: float = 2.0,
                 reduction: str = "mean",
                 ) -> None:
        super(FocalLoss, self).__init__()
        assert alpha is None or freq is None, "Only one between alpha and freq can be passed"
        if alpha is None and freq is None:
            self.alpha = 1.0
        elif alpha is None:
            self.alpha = freq * freq.shape[0] / torch.sum(freq)
        elif freq is None:
            self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction

    def forward(self,
                yhat, y
                ) -> torch.Tensor:
        if self.alpha != 1.0:
            self.alpha = self.alpha.to(y.device)
        pt = F.softmax(yhat, dim=-1)
        logpt = F.log_softmax(yhat, dim=-1)
        if self.alpha == 1.0:
            alpha = torch.ones_like(y, device=y.device).unsqueeze(dim=-1)
        else:
            alpha = self.alpha[y].unsqueeze(dim=-1)
        floss = F.nll_loss(alpha * ((1 - pt) ** self.gamma) * logpt, y, reduction="none")
        if self.reduction == "none":
            return floss
        elif self.reduction == "mean":
            return torch.mean(floss)
        elif self.reduction == "sum":
            return torch.sum(floss)



import io
import os
import sys
from shutil import rmtree
from setuptools import find_packages, setup, Command

# Meta-data
NAME = 'sifeng'
DESCRIPTION = 'The machine learning quant research framework.'
URL = 'https://github.com/zty200489/sifeng'
EMAIL = 'zty200489@gmail.com'
AUTHOR = 'Tianyuan Zhou'
REQUIRES_PYTHON = '>=3.8.0'

# Required Packages
REQUIRED = [
    "tqdm",
    "numpy",
    "pandas",
    "duckdb",
    "joblib",
    "requests",
    "colorama",
]

# Optional Packages
EXTRAS = {
}

# Global Settings
here = os.path.abspath(os.path.dirname(__file__))
# Load REAMDE.md
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load __version__.py
about = {}
project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
with open(os.path.join(here, project_slug, '__version__.py')) as f:
    exec(f.read(), about)

# Upload Commands
class UploadCommand(Command):
    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass
        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))
        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')
        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')
        sys.exit()


# Setup Process
setup(
    name = NAME,
    version = about['__version__'],
    description = DESCRIPTION,
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    author = AUTHOR,
    author_email = EMAIL,
    python_requires = REQUIRES_PYTHON,
    url = URL,
    packages = find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['sifeng'],
    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    install_requires = REQUIRED,
    extras_require = EXTRAS,
    include_package_data = True,
    license = 'GPL',
    classifiers = [
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Mathematics'
    ],
    # $ setup.py publish support.
    # cmdclass = {
    #     'upload': UploadCommand,
    # },
)



import torch
from torch import nn
from torch.nn.parameter import Parameter

__all__ = [
    "AbsoluteSinusoidalPE",
    "AbsoluteLearnablePE",
]

class AbsoluteSinusoidalPE(nn.Module):
    """The Absolute Sinusoidal Positional Encoding Module
    A. Vaswani et al., “Attention Is All You Need.” arXiv, Aug. 01, 2023. Accessed: Sep. 25, 2023.
    [Online]. Available: http://arxiv.org/abs/1706.03762

    Parameters
    ----------
    embed_dim: int
        the dimension of input embeddings
    dropout: float
        the probability of dropout
    max_len: int, dafault `128`
        the maximum length of positional encoding

    Input shape
    -----------
    [bsz * slen * embed_dim]
    """
    def __init__(self,
                 embed_dim: int,
                 dropout: float,
                 max_len: int = 128):
        super(AbsoluteSinusoidalPE, self).__init__()
        self.embed_dim = embed_dim
        self.dropout = dropout
        self.max_len = max_len

        self.dropout = nn.Dropout(p=dropout)
        self.Encoding = torch.empty(1, max_len, embed_dim)
        x = torch.arange(max_len, dtype=torch.float32).reshape(-1, 1) / torch.pow(max(10000, max_len), torch.arange(0, embed_dim, 2, dtype=torch.float32) / embed_dim)
        self.Encoding[:, :, 0::2] = torch.sin(x)
        self.Encoding[:, :, 1::2] = torch.cos(x)

    def forward(self,
                x: torch.Tensor, # [bsz * slen * embed_dim]
                ) -> torch.Tensor:
        bsz, slen, embed_dim = x.shape
        return x + self.dropout(self.Encoding[:, :slen, :])

class AbsoluteLearnablePE(nn.Module):
    """The Absolute Positional Encoding Module
    A. Dosovitskiy et al., “An Image is Worth 16x16 Words: Transformers for Image Recognition at
    Scale.” arXiv, Jun. 03, 2021. Accessed: Sep. 28, 2023. [Online]. Available:
    http://arxiv.org/abs/2010.11929

    Parameters
    ----------
    embed_dim: int
        the dimension of input embeddings
    dropout: float
        the probability of dropout
    max_len: int, dafault `128`
        the maximum length of positional encoding

    Input shape
    -----------
    [bsz * slen * embed_dim]
    """
    def __init__(self,
                 embed_dim: int,
                 dropout: float = 0.0,
                 max_len: int = 128,
                 ) -> None:
        super(AbsoluteLearnablePE, self).__init__()
        self.embed_dim = embed_dim
        self.dropout = dropout
        self.max_len = max_len

        self.dropout = nn.Dropout(p=dropout)
        self.Encoding = Parameter(torch.rand(1, max_len, embed_dim))

    def forward(self,
                x: torch.Tensor, # [bsz * slen * embed_dim]
                ) -> torch.Tensor:
        bsz, slen, embed_dim = x.shape
        return x + self.dropout(self.Encoding[:, :slen, :])



import torch, numpy as np
from torch import nn
from torch.nn.parameter import Parameter
from torch.nn.init import xavier_uniform_

from typing import Optional, Literal

__all__ = [
    "FeedForwardLayer",
    "MultiheadedSelfAttentionModule",
    "TransformerEncoderBlock",
]

class FeedForwardLayer(nn.Module):
    """The Feed-Forward Layer

    Parameters
    ----------
    embed_dim: int
        the dimension of input embeddings
    activation: torch.nn.Module, default torch.nn.ReLU
        the activation module used in the feed-forward sandwich
    elementwise_affine: bool, default `False`
        whether to enable element-wise affine in layer norm
    *args, **kwargs:
        Parameters for the activation function

    Input shape
    -----------
    [bsz * ?? * embed_dim]
    """
    def __init__(self,
                 embed_dim: int,
                 activation: nn.Module = nn.ReLU,
                 *args, **kwargs) -> None:
        super(FeedForwardLayer, self).__init__()
        self.sequential = nn.Sequential(
            nn.Linear(embed_dim, 4 * embed_dim),
            activation(*args, **kwargs),
            nn.Linear(4 * embed_dim, embed_dim)
        )

    def forward(self,
                x: torch.Tensor, # [bsz * ?? * embed_dim]
                ) -> torch.Tensor:
        return self.sequential(x)

class MultiheadedSelfAttentionModule(nn.Module):
    """The Multiheaded Self-Attention Module
    A. Vaswani et al., “Attention Is All You Need.” arXiv, Aug. 01, 2023. Accessed: Sep. 25, 2023.
    [Online]. Available: http://arxiv.org/abs/1706.03762

    Parameters
    ----------
    num_head: int
        number of self-attention heads
    embed_dim: int
        the dimension of input embeddings
    qk_dim: Optional[int], default `None`
        the dimension of qk vectors
    v_dim: Optional[int], default `None`
        the dimension of v vectors
    dropout: float, default `0.0`
        probability of dropout

    Input shape
    -----------
    [bsz * slen * embed_dim]
    """
    def __init__(self,
                 num_head: int,
                 embed_dim: int,
                 qk_dim: Optional[int] = None,
                 v_dim: Optional[int] = None,
                 dropout: float = 0.0,
                 ) -> None:
        super(MultiheadedSelfAttentionModule, self).__init__()
        self.num_head = num_head
        self.embed_dim = embed_dim
        self.qk_dim = qk_dim if qk_dim is not None else embed_dim
        self.v_dim = v_dim if v_dim is not None else embed_dim
        self.uniform_qkv_dim = (self.embed_dim == self.qk_dim and self.embed_dim == self.v_dim)

        if self.uniform_qkv_dim:
            self.head_dim = self.embed_dim // self.num_head
            assert self.head_dim * self.num_head == self.embed_dim, \
                "embed_dim not divisible by num_head"
            self.W_QKV = Parameter(torch.empty(self.embed_dim, self.embed_dim * 3))
        else:
            self.qk_head_dim = self.qk_dim // self.num_head
            self.v_head_dim = self.v_dim // self.num_head
            assert self.qk_head_dim * self.num_head == self.qk_dim, \
                "qk_dim not devisible by num_head"
            assert self.v_head_dim * self.num_head == self.v_dim, \
                "v_dim not divisible by num_head"
            self.W_QK = Parameter(torch.empty(self.embed_dim, self.qk_dim * 2))
            self.W_V = Parameter(torch.empty(self.embed_dim, self.v_dim))
        self.W_Out = Parameter(torch.empty(self.v_dim, self.embed_dim))
        self.softmax = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)

        self._init_params()

    def _init_params(self) -> None:
        if self.uniform_qkv_dim:
            xavier_uniform_(self.W_QKV)
        else:
            xavier_uniform_(self.W_QK)
            xavier_uniform_(self.W_V)
        xavier_uniform_(self.W_Out)

    # TODO: Functionize multiheaded-self attention module.
    def forward(self,
                x: torch.Tensor, # [bsz * slen * embed_dim]
                ) -> torch.Tensor:
        bsz, slen, embed_dim = x.shape
        if self.uniform_qkv_dim:
            qkv = torch.matmul(x, self.W_QKV).reshape(bsz, slen, self.num_head, self.head_dim, 3).transpose(1, 2) # [bsz * num_head * slen * head_dim * 3]
            q, k, v = qkv[:, :, :, :, 0], qkv[:, :, :, :, 1], qkv[:, :, :, :, 2] # [bsz * num_head * slen * head_dim]
        else:
            qk = torch.matmul(x, self.W_QK).reshape(bsz, slen, self.num_head, self.qk_head_dim, 2).transpose(1, 2) # [bsz * num_head * slen * qk_head_dim * 2]
            q, k = qk[:, :, :, :, 0], qk[:, :, :, :, 1] # [bsz * num_head * slen * qk_head_dim]
            v = torch.matmul(x, self.W_V).reshape(bsz, slen, self.num_head, self.v_head_dim).transpose(1, 2) # [bsz * num_head * slen * v_head_dim]
        attn = self.softmax(torch.matmul(q, k.transpose(2, 3)) / np.sqrt(self.qk_dim)) # [bsz * num_head * slen * slen]
        attn = self.dropout(attn) # [bsz * num_head * slen * slen]
        out = torch.matmul(attn, v).transpose(1, 2).reshape(bsz, slen, self.v_dim) # [bsz * slen * v_dim]
        out = torch.matmul(out, self.W_Out)
        return out

class TransformerEncoderBlock(nn.Module):
    """The Transformer Encoder Block
    - post layer-norm
      A. Vaswani et al., “Attention Is All You Need.” arXiv, Aug. 01, 2023. Accessed: Sep. 25,
      2023. [Online]. Available: http://arxiv.org/abs/1706.03762
    - pre layer-norm
      R. Xiong et al., “On Layer Normalization in the Transformer Architecture.” arXiv, Jun. 29,
      2020. Accessed: Sep. 28, 2023. [Online]. Available: http://arxiv.org/abs/2002.04745
    - sandwich layer-norm
      [UNKNOWN]
    - TODO: deepnorm
      H. Wang, S. Ma, L. Dong, S. Huang, D. Zhang, and F. Wei, “DeepNet: Scaling Transformers to
      1,000 Layers.” arXiv, Mar. 01, 2022. Accessed: Sep. 28, 2023. [Online]. Available:
      http://arxiv.org/abs/2203.00555

    Parameters
    ----------
    num_head: int
        the number of self-attention heads for the attention module
    embed_dim: int
        the dimension of input embeddings
    qk_dim: Optional[int], default `None`
        the dimension of qk vectors for the attention module
    v_dim: Optional[int], default `None`
        the dimension of v vectors for the attention module
    dropout: float, default `0.0`
        probability of dropout
    feed_forward: nn.Module, default `sifeng.dl.modules.FeedForwardLayer`
        the choice of feedforward layer
    layernorm_mode: Literal["post", "pre", "sanwich"], default `"pre"`
        The mode of layernorm
        - "post": layernorm happens after residual connection
        - "pre": layernorm happens before attention block, better when training without warmup
        - "sanwich": layernorm happens before and after attention block, better when training under FP16
    *args, **kwargs:
        parameters for your choice of feed forward network

    Input shape
    -----------
    [bsz * slen * embed_dim]
    """
    def __init__(self,
                 num_head: int,
                 embed_dim: int,
                 qk_dim: Optional[int] = None,
                 v_dim: Optional[int] = None,
                 dropout: float = 0.0,
                 feed_forward: nn.Module = FeedForwardLayer,
                 layernorm_mode: Literal["post", "pre", "sanwich"] = "pre",
                 *args, **kwargs) -> None:
        super(TransformerEncoderBlock, self).__init__()
        self.layernorm_mode = layernorm_mode
        self.layernorm = nn.LayerNorm(embed_dim, elementwise_affine=False)
        self.feed_forward = feed_forward(embed_dim=embed_dim, *args, **kwargs)
        self.attention = MultiheadedSelfAttentionModule(num_head, embed_dim, qk_dim=qk_dim, v_dim=v_dim, dropout=dropout)
        self.dropout = nn.Dropout(dropout)

    def forward(self,
                x: torch.Tensor, # [batch * slen * embed_dim]
                ) -> torch.Tensor:
        if self.layernorm_mode == "post":
            x = self.layernorm(x + self.dropout(self.attention(x))) # [batch * slen * embed_dim]
            x = self.layernorm(x + self.dropout(self.feed_forward(x))) # [batch * slen * embed_dim]
        elif self.layernorm_mode == "pre":
            x = x + self.dropout(self.attention(self.layernorm(x))) # [batch * slen * embed_dim]
            x = x + self.dropout(self.feed_forward(self.layernorm(x))) # [batch * slen * embed_dim]
        elif self.layernorm_mode == "sandwich":
            x = x + self.dropout(self.layernorm(self.attention(self.layernorm(x)))) # [batch * slen * embed_dim]
            x = x + self.dropout(self.layernorm(self.feed_forward(self.layernorm(x)))) # [batch * slen * embed_dim]
        return x
